from functools import wraps
from time import perf_counter as counter
from typing import Any, Callable, ParamSpec, Protocol, overload

from horology.tformatter import UnitType, rescale_time

P = ParamSpec('P')


class CallableWithInterval(Protocol[P]):
    """Protocol to represent a callable with interval attribute.

    References
    ----------
    [PEP 612](https://peps.python.org/pep-0612/)
    """
    interval: float
    __call__: Callable[P, Any]
    __name__: str


@overload
def timed(f: Callable[P, Any]) -> CallableWithInterval[P]: ...  # Bare decorator usage


@overload
def timed(
        *,
        name: str | None = None,
        unit: UnitType = 'auto',
        print_fn: Callable[..., Any] | None = print
) -> Callable[[Callable[P, Any]], CallableWithInterval[P]]: ...  # Decorator with arguments


def timed(
        f: Callable[P, Any] | None = None,
        *,
        name: str | None = None,
        unit: UnitType = 'auto',
        print_fn: Callable[..., Any] | None = print):
    """Decorator that prints time of execution of the decorated function

    Parameters
    ----------
    f: Callable
        The function which execution time should be measured.
    name: str or None, optional
        String that should be printed as the function name. By default,
        the f.__name__ proceeded by a colon and space is used. See
        examples below.
    unit: {'auto', 'ns', 'us', 'ms', 's', 'min', 'h', 'd'}
        Time unit used to print elapsed time. Use 'a' or 'auto' for
        automatic time adjustment (default).
    print_fn: Callable or None, optional
        Function that is called to print the time elapsed. Use `None` to
        disable printing anything. You can provide e.g. `logger.info`.
        By default, the built-in `print` function is used.

    Attributes
    ----------
    interval: float
        Time elapsed by the function in seconds. Can be used to get the
        time programmatically after the execution of f.

    Returns
    -------
    Callable
        Decorated function `f`.

    Examples
    --------
    Basic usage
        ```
        @timed
        def foo():
            ...
        foo() # prints 'foo: 5.12 ms'
        ```

    Change default name
        ```
        @timed(name='bar elapsed ')
        def bar():
            ...
        bar() # prints 'bar elapsed 2.56 ms'
        ```

    Change default units
        ```
        @timed(unit='ns')
        def baz():
            ...
        baz() # prints 'baz: 3.28e+04 ns'
        ```

    Suppress printing and use the attribute `interval` to get the time
    elapsed
        ```
        @timed(print_fn=None)
        def qux():
            ...
        qux() # prints nothing
        print(qux.interval)
        ```

    """

    def decorator(_f):
        @wraps(_f)
        def wrapped(*args, **kwargs):
            start = counter()
            exception = None
            try:
                return_value = _f(*args, **kwargs)
            except Exception as e:
                exception = e
            finally:
                interval = counter() - start
                wrapped.interval = interval

            if print_fn is not None:
                nonlocal name
                name = _f.__name__ + ': ' if name is None else name
                t, u = rescale_time(interval, unit=unit)
                print_str = f'{name}{t:.3g} {u}'
                if exception is not None:
                    print_str += ' (failed)'
                print_fn(print_str)

            if exception is not None:
                raise exception

            return return_value

        return wrapped

    if f is None:  # used with ()
        return decorator
    else:  # used without ()
        return decorator(f)



from __future__ import annotations

from statistics import mean, median, stdev
from time import perf_counter as counter
from typing import Any, Callable, Iterable

from horology.tformatter import UnitType, rescale_time


class Timed:
    """ Wrapper to an iterable that measures time of each iteration

    Parameters
    ----------
    iterable: Iterable
        Object that should we wrapped.
    unit: str, optional
        Time unit used to print elapsed time. Possible values:
         ['ns', 'us', 'ms', 's', 'min', 'h', 'd']. Use 'a' or 'auto'
         for automatic time adjustment (default).
    iteration_print_fn: Callable, optional
        Function that is called after each iteration to print time
        of that iteration. Use `None` to disable printing after each
        iteration. You can provide e.g. `logger.debug`. By default,
        the built-in `print` function is used.
    summary_print_fn: Callable, optional
        Function that is called to print the summary. Use `None` to
        disable printing the summary. You can provide e.g.
        `logger.info`. By default, the built-in `print` function is used.

    Attributes
    ----------
    num_iterations: int
        How many iteration were executed.
    total: float
        Total time elapsed in seconds.

    Example
    -------
    Basic usage
        ```
        from horology import Timed
        animals = ['cat', 'dog', 'crocodile']
        for x in Timed(animals):
            feed(x)
        ```

        Possible result:
        ```
        iteration    1: 12.0 s
        iteration    2: 8.00 s
        iteration    3: 100 s

        total 3 iterations in 120 s
        min/median/max: 8.00/12.0/100 s
        average (std): 40.0 (52.0) s
        ```
    """

    def __init__(
            self,
            iterable: Iterable,
            *,
            unit: UnitType = 'a',
            iteration_print_fn: Callable[..., Any] | None = print,
            summary_print_fn: Callable[..., Any] | None = print
    ) -> None:

        self.iterable = iterable
        self.unit = unit
        self.iteration_print_fn = iteration_print_fn or (lambda _: None)
        self.summary_print_fn = summary_print_fn or (lambda _: None)

        self.intervals: list[float] = []
        self._start: float | None = None
        self._last: float | None = None

    def __iter__(self) -> Timed:
        self._start = counter()
        self.iterable = iter(self.iterable)
        return self

    def __next__(self):
        try:
            now = counter()
            if self._last is not None:
                interval = now - self._last
                self.intervals.append(interval)
                t, u = rescale_time(interval, self.unit)
                self.iteration_print_fn(f'iteration {self.num_iterations:4}: {t:.3g} {u}')

            self._last = now

            return next(self.iterable)

        except StopIteration:
            self.print_summary()
            raise StopIteration

    @property
    def num_iterations(self) -> int:
        return len(self.intervals)

    @property
    def n(self) -> int:
        "Deprecated"
        return self.num_iterations

    @property
    def total(self) -> float:
        try:
            return self._last - self._start  # type: ignore
        except TypeError:
            return 0

    def print_summary(self) -> None:
        """ Print statistics of times elapsed in each iteration

        It is called automatically when the iteration over the iterable
        finishes.

        Use `summary_print_fn` argument in the constructor to control
        if and where the summary is printed.

        """
        # Leave an empty line if iterations and summary are printed to
        # the same output.
        if self.iteration_print_fn == self.summary_print_fn:
            print_str = '\n'
        else:
            print_str = ''

        if self.num_iterations == 0:
            print_str = 'no iterations'
        elif self.num_iterations == 1:
            t, u = rescale_time(self.intervals[0], unit=self.unit)
            print_str += f'one iteration: {t:.3g} {u}'
        else:
            t_total, u_total = rescale_time(self.total, self.unit)

            t_median, u = rescale_time(median(self.intervals), self.unit)
            # For clarity, all values are shown using the same unit.
            t_min, _ = rescale_time(min(self.intervals), u)
            t_mean, _ = rescale_time(mean(self.intervals), u)
            t_max, _ = rescale_time(max(self.intervals), u)
            t_std, _ = rescale_time(stdev(self.intervals), u)

            print_str += f'total {self.num_iterations} iterations '
            print_str += f'in {t_total:.3g} {u_total}\n'
            print_str += f'min/median/max: ' \
                         f'{t_min:.3g}' \
                         f'/{t_median:.3g}' \
                         f'/{t_max:.3g} {u}\n'
            print_str += f'average (std): ' \
                         f'{t_mean:.3g} ' \
                         f'({t_std:.3g}) {u}'

        self.summary_print_fn(print_str)



from math import inf
from typing import Literal, NamedTuple, cast

UnitType = Literal['a', 'auto', 'ns', 'us', 'ms', 's', 'min', 'h', 'd']


class Unit(NamedTuple):
    name: UnitType
    scale: float
    limit: float


UNITS = [
    Unit('ns', 10 ** -9, 10 ** -6),
    Unit('us', 10 ** -6, 10 ** -3),
    Unit('ms', 10 ** -3, 1),
    Unit('s', 1, 10 ** 3),
    Unit('min', 60, 6 * 10 ** 4),
    Unit('h', 3600, 36 * 10 ** 5),
    Unit('d', 3600 * 24, inf),
]


def rescale_time(
        interval: float,
        unit: UnitType,
) -> tuple[float, UnitType]:
    """Rescale the time interval using the provided unit

    Parameters
    ----------
    interval
        Time interval to be rescaled, in seconds.
    unit
        Time unit to which `interval` should be rescaled. Use 'a' or
        'auto' for automatic time adjustment.

    Returns
    -------
    float
        Time interval in new units.
    str
        Convenient unit found automatically if input unit was 'auto'.
        Otherwise, provided `unit` is returned unchanged.

    Examples
    --------
    >>> rescale_time(0.421, 'us')
    (421000.0, 'us')
    >>> rescale_time(150, 'min')
    (2.5, 'min')
    >>> rescale_time(0.911, 'auto')
    (911.0, 'ms')

    Raises
    ------
    ValueError
        If the unit provided is unknown.

    """
    unit = cast(UnitType, unit.lower())

    for u in UNITS:
        if unit == u.name or (unit in ('a', 'auto') and interval < u.limit):
            return interval / u.scale, u.name

    raise ValueError(f"Unknown unit: {unit}. Use one of the following: "
                     f"{[x.name for x in UNITS]} or 'auto'")
